# orm_models.py
# -*- coding: utf-8 -*-

"""
ORM (Object-Relational-Mapping) objects which serve for database purposes
"""

import sqlalchemy as sa
from sqlalchemy.orm import Session
from sqlalchemy.orm.attributes import InstrumentedAttribute


# ______________________________________________________________________________________________________________________


Base = sa.orm.declarative_base()


class BaseClass:
    """
    Base class which implements general functionality
    """

    @classmethod
    def set_table_attribute(cls, attr_name: str, attr_value: object) -> None:
        """
        sets the attribute for the table, e.g. 'schema' = 'job' or 'name' = 'job_exec'

        Args:
            attr_name: the name of the table attribute
            attr_value: the value to be set for the attribute

        Returns:
            None
        """

        setattr(cls.__table__, attr_name, attr_value)
        return None

    @classmethod
    def create_table(cls, engine: sa.engine.Engine, recreate: bool = False) -> None:
        """
        creates a job execution information table. Only use once!

        Args:
            engine: the sqlalchemy engine to the database
            recreate: if true drop the existing table and recreate it

        Returns:
            Creates the specified table and returns None
        """

        if recreate:
            cls.__table__.drop(engine, checkfirst=True)
        cls.__table__.create(engine, checkfirst=True)

        return None

    @classmethod
    def select_all(cls, engine: sa.engine.Engine) -> list[object]:
        """
        retrieves all rows as orm instances to a list

        Args:
            engine: the sqlalchemy engine used to connect to the database

        Returns:
            list of all object instances
        """
        with Session(engine) as session:
            session.expire_on_commit = False  # keep the instance accessible after session closes
            stmt = sa.select(cls)
            rows = session.execute(stmt).all()
            session.commit()
            rows = [r[0] for r in rows]

        return rows

    @classmethod
    def select_filtered(cls, engine: sa.engine.Engine, verbose: bool = False, **kwargs) -> list[object]:
        """
        retrieves filtered rows as orm instances to a list

        Args:
            engine: the sqlalchemy engine used to connect to the database
            verbose: if true print the generated sql statement
            **kwargs: the key value pairs corresponding to table colum and colum value to filter

        Returns:
            filtered list of object instances
        """
        with Session(engine) as session:
            session.expire_on_commit = False  # keep the instance accessible after session closes
            stmt = sa.select(cls).filter_by(**kwargs)
            if verbose:
                print(stmt)
            rows = session.execute(stmt).all()
            session.commit()
            rows = [r[0] for r in rows]

        return rows

    def add(self, engine: sa.engine.Engine) -> object:
        """
        writes the current instance to the database

        Args:
            engine: the sqlalchemy engine used to connect to the database

        Returns:
            self
        """

        with Session(engine) as session:
            session.expire_on_commit = False  # keep the instance accessible after session closes
            if self.id is None:
                session.add(self)
            session.commit()
        return self

    def update(self, engine: sa.engine.Engine) -> None:
        """
       updates the current instance to the database

        Args:
            engine: the sqlalchemy engine used to connect to the database

        Returns:
            None
        """

        with Session(engine) as session:
            session.expire_on_commit = False  # keep the instance accessible after session closes
            mapped_values = {}
            primary_keys = [element.name for element in self.__class__.__table__.primary_key]
            for key, value in self.__class__.__dict__.items():
                if isinstance(value, InstrumentedAttribute) and key not in primary_keys:
                    mapped_values[key] = getattr(self, key)

            session.query(self.__class__).filter(self.__class__.id == self.id).update(mapped_values)
            session.commit()
        return None


# ______________________________________________________________________________________________________________________


class JobExec(Base, BaseClass):
    """
    ORM model for the job_execution table
    """

    __tablename__ = 'job_execution'
    __table_args__ = {'schema': None}

    id = sa.Column(sa.Integer, sa.Identity(start=1, cycle=True), primary_key=True)
    name = sa.Column(sa.String(255), nullable=False)
    start_time = sa.Column(sa.DateTime, nullable=False)
    end_time = sa.Column(sa.DateTime, nullable=True)
    status = sa.Column(sa.String(255), nullable=True)
    running = sa.Column(sa.Boolean, nullable=False)
    executor = sa.Column(sa.String(31), nullable=False)
    exception = sa.Column(sa.String(255), nullable=True)
    log_file = sa.Column(sa.String(255), nullable=True)

    def __repr__(self):
        return f'JobExec(id={self.id}, name={self.name}, start_time={self.start_time}, status={self.status})'


# ______________________________________________________________________________________________________________________


class JobSchedule(Base, BaseClass):
    """
    ORM model for the job_schedule table
    """

    __tablename__ = 'job_schedule'
    __table_args__ = {'schema': None}

    id = sa.Column(sa.Integer, sa.Identity(start=1, cycle=True), primary_key=True)
    name = sa.Column(sa.String(255), nullable=False)
    location = sa.Column(sa.String(255), nullable=False)
    args = sa.Column(sa.String(255), nullable=False)
    env = sa.Column(sa.String(255), nullable=False)
    schedule = sa.Column(sa.String(255), nullable=False)
    priority = sa.Column(sa.SmallInteger, nullable=True)
    active = sa.Column(sa.Boolean, default=True, nullable=False)

    def __repr__(self):
        schedule = zip(self.schedule.split(), ['minute', 'hour', 'day', 'month', 'weekday'])
        schedule = ', '.join([f'{j}={i}' for i, j in schedule])
        return f'JobSchedule(id={self.id}, name={self.name}, args={self.args}, schedule=[{schedule}])'


# ______________________________________________________________________________________________________________________



# util_functions.py
# -*- coding: utf-8 -*-

"""
Functions which serve for general purposes
"""

from warnings import warn

import os
import sys
import psutil
from pathlib import Path

import re

import typing
import inspect

from colorama import (
    Fore,
    Style,
)

import logging
import warnings
import traceback

from datetime import datetime, timedelta
from dateutil import parser as dateutil_parser

import pandas as pd


# ______________________________________________________________________________________________________________________


def replace_none(obj: typing.Any, replacement: typing.Any) -> typing.Any:
    """
    mimics the sql isnull() function

    Args:
        obj: input object
        replacement: replacement object if input object is None

    Returns:
        replacement if obj is None else obj
    """
    if obj is None:
        return replacement
    return obj


# ______________________________________________________________________________________________________________________


def print_yellow(*args, **kwargs) -> None:
    """
    colors the text yellow which will be printed

    Args:
        *args: arguments with are passed on to the builtin print function
        **kwargs: keyword arguments with are passed on to the builtin print function

    Returns:
        None, but prints the text in yellow
    """
    sep = kwargs.pop('sep', ' ')
    original_string = sep.join(str(stmt) for stmt in args)
    print(f'{Fore.YELLOW}{original_string}{Style.RESET_ALL}', **kwargs)


# ______________________________________________________________________________________________________________________


def input_yellow(prompt: str) -> str:
    """
    colors the text yellow which will be printed by the prompt for input

    Args:
        prompt: the argument which will be passed on to the builtin function input

    Returns:
        the user input
    """
    return input(f'{Fore.YELLOW}{prompt}{Style.RESET_ALL}')


# ______________________________________________________________________________________________________________________


def determine_main_script_path() -> str:
    """
    Determines the path of the main script using psutil.Process().cmdline()

    Returns:
        a pathlib.Path object with the path to the main script
    """
    args = psutil.Process().cmdline()
    if len(args) > 1:
        path = args[1]
    else:
        path = args[0]
    return path


def runs_in_repl_mode() -> bool:
    """
    Determines if the execution is in repl (read-evaluate-print-loop) mode.

    Returns:
        True if reple else False
    """
    return hasattr(sys, 'ps1') or Path(determine_main_script_path()).name == 'pydevconsole.py'


# ______________________________________________________________________________________________________________________


def add_to_namespace(name: str, value: object, namespace: dict) -> None:
    """
    adds a variable to the namespace

    Args:
        name: the name of the variable
        value: the value of the variable
        namespace: the namespace where it should be added

    Returns:
        None
    """

    if name in namespace:
        warnings.warn(
            f'Function <{inspect.currentframe().f_code.co_name}>: '
            f'overwriting <{name}> which already existed in the namespace!'
        )
    namespace[name] = value
    return None


# ______________________________________________________________________________________________________________________


def check_if_in_argv(arg, argument) -> bool:
    """
    checks if -arg or --argument is in sys.argv

    Args:
        arg: the name of the argument in short notation
        argument: the name of the argument in verbose notation

    Returns:
        True if the argument is part of argv, else False
    """
    import sys
    return f'-{arg}' in sys.argv or f'--{argument}' in sys.argv


# ______________________________________________________________________________________________________________________


def convert_to_datetime(s: str) -> datetime:
    """
    convert a string input to python native datetime object

    Args:
        s: input string

    Returns:
        datetime object or pd.NaT if input is pd.isna()
    """
    if pd.isna(s):
        return pd.NaT
    elif isinstance(s, pd.Timestamp):
        return s.to_pydatetime()
    elif isinstance(s, datetime):
        return s
    return dateutil_parser.parse(s)


# ______________________________________________________________________________________________________________________


def parse_bool(arg: str) -> bool:
    """
    parse a script argument to a boolean value

    Args:
        arg: the input argument

    Returns:
        Boolean value of arg
    """
    inp = str(arg).lower()
    if 'true'.startswith(inp):
        return True
    elif 'false'.startswith(inp):
        return False
    else:
        raise ValueError(f'arg needs to be one of [ True | False ], but arg was {str(arg)}')


# ______________________________________________________________________________________________________________________


def input_prompt(
        name: str,
        message: str = None,
        choices: tuple = (None, ),
        multi: bool = False,
        default: object = None,
        enum: bool = False
) -> str | list[str]:
    """
    wrapper for pythons input() with choices, default value and continuous prompting if an invalid input was supplied

    Args:
        name: the name of the variable
        message: the custom message to be printed.
                    If not None, name is omitted, else `please set the {name}:` will be printed
        choices: the allowed values for input. If None, anything can be input
        multi: whether to allow multiple selection (comma separated)
        default: the default value. If None, the user will continue to be prompted
        enum: enumerate the choices and allow for numerical input

    Returns:
        user input
    """
    print()
    if message is not None:
        print_yellow(f'{message}:')
    else:
        print_yellow(f'please set the {name}:')

    inp = None
    if choices == (None, ) and default is None:
        inp = input().strip()
        if multi:
            inp = [el.strip() for el in inp.split(',')]
    elif choices == (None, ) and default is not None:
        inp = input_yellow(f'\t-> defaults to: {default}').strip()
        inp = default if inp == '' else inp
        if multi:
            inp = [el.strip() for el in inp.split(',')]
    elif not enum:
        set_inp = set()
        while inp is None or not set_inp.issubset(set(choices)):
            inp = input_yellow(
                f'\t-> choose between [{", ".join(str(e) for e in choices)}], defaults to: {default} '
            ).strip()
            inp = default if inp == '' else inp
            if multi:
                inp = [el.strip() for el in inp.split(',')]
            set_inp = set(inp) if multi else {inp}
    else:
        available_choices = {i: item for i, item in enumerate(choices, start=1)}
        print_yellow('  choose from', end='')
        print_yellow('\t', *available_choices.items(), sep='\n\t')
        print_yellow(f'  default will be {default}')
        set_inp = set()
        while inp is None or not (inp == ''
                                  or set_inp.issubset(set(available_choices.values()))
                                  or set_inp.issubset(set([str(i) for i in available_choices.keys()]))):
            inp = input_yellow('\t-> enter the number or value of your choice ')
            inp = default if inp == '' else inp
            if multi:
                inp = [el.strip() for el in inp.split(',')]
            set_inp = set(inp) if multi else {inp}

        if not multi and inp.isdigit():
            inp = available_choices[int(inp)]
        elif multi and all(el.isdigit() for el in inp):
            inp = [available_choices[int(el)] for el in inp]

    print_yellow(f'user input: {inp}')
    print()

    return inp


# ______________________________________________________________________________________________________________________


def prompt_file_name(open_or_save: str = 'open', gui: bool = True) -> str:
    """
    Prompts the user to select a file.

    Args:
        open_or_save: whether it should prompt for an existing file (open) or to save a new file (save)
        gui: whether to use tkinter gui or just plain input

    Returns:
        the path to the file
    """
    file_path = None
    if open_or_save not in ('open', 'save'):
        raise ValueError(f"Argument open_or_save must be 'open' or 'save'. Input was {open_or_save}")

    if gui:
        import tkinter as tk
        window = tk.Tk()
        window.wm_attributes('-topmost', 1)
        window.withdraw()
        if open_or_save == 'open':
            from tkinter.filedialog import askopenfilename
            file_path = askopenfilename(parent=window, title='please select the file name')
        elif open_or_save == 'save':
            from tkinter.filedialog import asksaveasfilename
            file_path = asksaveasfilename(parent=window, title='please set the file name to be saved')
    else:
        file_path = str(input_prompt(name='path/to/file'))
        if open_or_save == 'open' and not Path(file_path).exists():
            raise ValueError(f'{file_path} does not exist!')
        if open_or_save == 'open' and not Path(file_path).is_file():
            raise ValueError(f'{file_path} is not a file!')

    return file_path


# ______________________________________________________________________________________________________________________


def determine_default_value_for_argparse(
        repl: bool,
        arg_name: tuple,
        arg_base_type: str = 'str',
        choices: tuple = (None, ),
        default: object = None,
        enum: bool = False,
) -> object:
    """
    only use for argparse.add_argument default value
    this function determines the default value based on execution mode and user input

    Args:
        repl: whether the script is executed in interactive repl (read-evaluate-print-loop) mode
        arg_name: the name of the arguments, e.g. ('-arg', '--argument')
        arg_base_type: the base type of the input (if arg is a list, then arg_base_type is the type of the elements)
        choices: the allowed values for input. If None, anything can be input
        default: the default value. If None, the user will continue to be prompted
        enum: enumerate the choices and allow for numerical input

    Returns:
        the determined default value
    """

    arg_name = tuple(arg.lstrip('-') for arg in arg_name)
    determined_default = default
    if repl:
        determined_default = input_prompt(arg_name[-1], choices=choices, default=default, enum=enum)
        if arg_base_type == 'int' and len(determined_default) > 0:
            determined_default = [int(el) for el in determined_default.split()]
        elif arg_base_type == 'float' and len(determined_default) > 0:
            determined_default = [float(el) for el in determined_default.split()]
        elif arg_base_type == 'bool' and len(determined_default) > 0:
            determined_default = [bool(el) for el in determined_default.split()]
    else:
        if not check_if_in_argv(*arg_name):
            if default is not None:
                determined_default = default
            else:
                determined_default = input_prompt(arg_name[-1], choices=choices, default=default, enum=enum)

    return determined_default


# ______________________________________________________________________________________________________________________


def create_logger(name: str, log_file_path: str | Path = None) -> logging.Logger:
    """
    logger factory function

    Args:
        name: the name of the logger
        log_file_path: the path to the logging file, if None, then no file handler will be added

    Returns:
        an instance of a logger
    """

    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)

    logging_console_formatter = logging.Formatter('%(levelname)-8s | %(message)s')
    logging_console_handler = logging.StreamHandler()
    logging_console_handler.setLevel(logging.DEBUG)
    logging_console_handler.setFormatter(logging_console_formatter)
    logger.addHandler(logging_console_handler)

    if log_file_path is not None:
        if isinstance(log_file_path, str):
            log_file_path = Path(log_file_path)
        if not log_file_path.exists():
            log_file_path.parent.mkdir(parents=True, exist_ok=True)
            log_file_path.touch(exist_ok=True)
        logging_file_formatter = logging.Formatter(
            '%(asctime)s | %(levelname)-8s | module: %(module)-25s | %(message)s'
        )
        logging_file_handler = logging.FileHandler(log_file_path)
        logging_file_handler.setLevel(logging.INFO)
        logging_file_handler.setFormatter(logging_file_formatter)
        logger.addHandler(logging_file_handler)

    return logger


# ______________________________________________________________________________________________________________________


def format_email_error_message(exception: Exception, project: str, wd: Path, log_file_path: Path) -> tuple:
    """
    prepares a formatted subject and body for the e_mail message

    Args:
        exception: the raised exception from the try block
        project: the project name
        wd: the working directory
        log_file_path: the Path object to the log file

    Returns:
        tuple of strings (e_mail_subject, e_mail_message)
    """

    tb = traceback.TracebackException.from_exception(exception)
    traceback_error = ''.join(tb.format())
    error = tb.exc_type.__name__

    e_mail_subject = f'Error {error} in {project}'
    with open(wd / 'templates' / 'email' / 'error_message.html', 'r', encoding='utf-8') as f:
        e_mail_message = f.read()
    e_mail_message = e_mail_message.replace('{{ project }}', project)
    e_mail_message = e_mail_message.replace('{{ datetime_now }}', datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
    e_mail_message = e_mail_message.replace('{{ error }}', error)
    e_mail_message = e_mail_message.replace('{{ traceback_error }}', traceback_error)
    e_mail_message = e_mail_message.replace('{{ log_file_path }}', str(log_file_path.resolve()))

    return e_mail_subject, e_mail_message

# ______________________________________________________________________________________________________________________


def chunker(seq: typing.Sequence, size: int) -> typing.Generator:
    """
    creates a generator object of the sequence in fixed size chunks

    Args:
        seq: the input sequence
        size: the chunk size (last chunk will only contain the leftover)

    Returns:
        generator object with chunked sequence elements of the specified size
    """

    return (seq[pos:pos + size] for pos in range(0, len(seq), size))


# ______________________________________________________________________________________________________________________


def split_text(text: str, n_chars: int = 50):
    """
    inserts `\n` into a long text at the first whitespace before `n_chars`.

    Args:
        text: the input text
        n_chars: the maximum number of chars per line

    Returns:
        text with `\n` inserted
    """

    result = ''
    start = 0
    while True:
        # loop exit condition
        if len(text[start:]) <= n_chars:
            result += text[start:]
            break

        # set the scope to analyse
        sub_text = text[start:start+n_chars]

        # first find the nearest whitespace to the end of text[:n_chars]
        ws_pos = sub_text.rfind(' ')

        # assembling the result string
        result += text[start:start+ws_pos] + '\n'

        # setting the new start point
        start = start + ws_pos + 1

    return result if result != '' else text


# ______________________________________________________________________________________________________________________


def last_day_of_month(any_date: datetime.date) -> datetime.date:
    """
    calculates the last day of the month for the supplied date

    Args:
        any_date: the input date

    Returns:
        date of the last day from that month
    """

    next_month = any_date.replace(day=28) + timedelta(days=4)  # this will never fail
    return next_month - timedelta(days=next_month.day)


# ______________________________________________________________________________________________________________________


def date_to_integer(any_date: datetime.date) -> int:
    """
    converts any date object to integer

    Args:
        any_date: the date which should be converted to int

    Returns:
        date as int
    """
    return (10000 * any_date.year) + (100 * any_date.month) + any_date.day


# ______________________________________________________________________________________________________________________


def display_formatted_time(seconds: int, granularity: str = 's') -> str:
    """
    outputs a string of the time passed in second, minutes, hours, days, weeks

    Args:
        seconds: the time passed in seconds
        granularity: controls how much unit details should be shown. Default 's' shows all non-zero units

    Returns:
        formatted string with elapsed time
    """

    result = []

    d_granularity = {s: i for s, i in zip(['w', 'd', 'h', 'm', 's'], [1, 2, 3, 4, 5])}
    granularity = d_granularity[granularity]

    intervals = (
        ('w', 604800),  # 60 * 60 * 24 * 7
        ('d', 86400),  # 60 * 60 * 24
        ('h', 3600),  # 60 * 60
        ('m', 60),
        ('s', 1),
    )

    seconds = int(round(seconds))
    if seconds == 0:
        return '0 s'

    for name, count in intervals:
        value = seconds // count
        if value:
            seconds -= value * count
            result.append(f'{value} {name}')

    return ', '.join(result[:granularity])


# ______________________________________________________________________________________________________________________


def clean_umlauts(s: str | list) -> str | list:
    """
    replaces some German, French, Slavic umlauts to plain english characters

    Args:
        s: the input string or list of strings

    Returns:
        input object with replaced umlauts
    """

    warn('deprecation warning: use unidecode instead. Install it with >>> pip install unidecode')
    d_trans = {
        'Ä': 'Ae',
        'Ü': 'Ue',
        'Ö': 'Oe',
        'É': 'E',
        'È': 'E',
        'À': 'A',
        'Ó': 'O',
        'Â': 'A',
        'Ê': 'e',
        'Ç': 'C',
        'Ć': 'C',
        'Č': 'C',
        'Ë': 'E',

        'ä': 'ae',
        'ü': 'ue',
        'ö': 'oe',
        'é': 'e',
        'è': 'e',
        'à': 'a',
        'ó': 'o',
        'â': 'a',
        'ê': 'e',
        'ç': 'c',
        'ć': 'c',
        'č': 'c',
        'ë': 'e',
    }

    for key, val in d_trans.items():
        if isinstance(s, str):
            s = s.replace(key, val)

        elif isinstance(s, list):
            s = [el.replace(key, val) for el in s]

    return s


# ______________________________________________________________________________________________________________________


def calc_equidistant_weights(n: int) -> list:
    """
    calculates the equidistant weights for n inputs. Sum of all weights equals 1

    Args:
        n: the number of weights needed

    Returns:
        list of equidistant weights
    """

    if n < 2:
        print('n needs to be at least 2')
        return [1]

    unidist = ((1 / n) / 2) / (n // 2)

    if n % 2 == 0:
        rng = [i for i in range(-(n // 2), n // 2 + 1) if i != 0]
    else:
        rng = [i for i in range(-(n // 2), n // 2 + 1)]

    return [round(1 / n + i * unidist, 10) for i in reversed(rng)]


# ______________________________________________________________________________________________________________________


yaml_env_resolver = re.compile(r'\${.+}')


def yaml_env_constructor(loader, node):
    value = node.value
    match = yaml_env_resolver.match(value)
    env_var = match.group()[2:-1]
    return os.environ.get(env_var) + value[match.end():]


# ______________________________________________________________________________________________________________________



# util_gui_classes.py
# -*- coding: utf-8 -*-

"""
Classes which serve for gui applications.
"""

from typing import Any

from _tkinter import TclError
import tkinter
import tkinter.messagebox
import customtkinter

from .util_functions import split_text


# ______________________________________________________________________________________________________________________


customtkinter.set_appearance_mode('System')  # Modes: 'System' (standard), 'Dark', 'Light'
customtkinter.set_default_color_theme('blue')  # Themes: 'blue' (standard), 'green', 'dark-blue'


# ______________________________________________________________________________________________________________________


class GuiPromptYesNo(customtkinter.CTk):
    """
    Creates a yes / no gui based prompt with default value and countdown functionality.
    The user input will be stored in:
    >>> instance.answer
    """
    WIDTH = 500
    HEIGHT = 200

    def __init__(self, question: str, default_value: str = 'no', countdown_seconds: int = 0):
        super().__init__()
        self.terminated = False

        self.title('input required')
        self.geometry(f'{self.__class__.WIDTH}x{self.__class__.HEIGHT}')
        self.protocol('WM_DELETE_WINDOW', self.on_closing)  # call .on_closing() when app gets closed
        self.resizable(False, False)

        if len(question) > 50:
            question = split_text(text=question, n_chars=50)
        self.question = question
        self.answer = None
        self.default_value = default_value
        self.countdown_seconds = countdown_seconds
        self.remaining_seconds = countdown_seconds

        # ============ create top-level-frames ============

        # configure grid layout (4x1)
        self.equal_weighted_grid(self, 4, 1)
        self.grid_rowconfigure(0, minsize=10)
        self.grid_rowconfigure(3, minsize=10)

        self.frame_label = customtkinter.CTkFrame(master=self, corner_radius=10)
        self.frame_label.grid(row=1, column=0)

        self.frame_buttons = customtkinter.CTkFrame(master=self, corner_radius=0, fg_color=None)
        self.frame_buttons.grid(row=2, column=0)

        # ============ design frame_label ============

        # configure grid layout (5x4)
        self.equal_weighted_grid(self.frame_label, 5, 4)
        self.frame_label.grid_rowconfigure(0, minsize=10)
        self.frame_label.grid_rowconfigure(2, minsize=10)
        self.frame_label.grid_rowconfigure(5, minsize=10)

        self.label_question = customtkinter.CTkLabel(
            master=self.frame_label,
            text=self.question,
            font=('Consolas', 12),
        )
        self.label_question.grid(row=1, column=0, columnspan=4, pady=5, padx=10)

        self.label_default_value = customtkinter.CTkLabel(
            master=self.frame_label,
            text='default value: ',
            font=('Consolas', 12),
        )
        self.label_default_value.grid(row=3, column=0, pady=5, padx=10)

        self.entry_default_value = customtkinter.CTkEntry(
            master=self.frame_label,
            width=40,
            justify='center',
            placeholder_text=self.default_value,
            state='disabled',
            textvariable=tkinter.StringVar(value=self.default_value),
            font=('Consolas', 12),
        )
        self.entry_default_value.grid(row=3, column=1, pady=5, padx=10)

        if countdown_seconds > 0:
            self.label_timer = customtkinter.CTkLabel(
                master=self.frame_label,
                text='timer [s]: ',
                font=('Consolas', 12),
            )
            self.label_timer.grid(row=3, column=2, pady=5, padx=10)

            self.entry_timer = customtkinter.CTkEntry(
                master=self.frame_label,
                width=40,
                justify='center',
                state='disabled',
                textvariable=tkinter.StringVar(value=str(self.remaining_seconds)),
                placeholder_text=str(self.remaining_seconds),
                font=('Consolas', 12),
            )
            self.entry_timer.grid(row=3, column=3, pady=5, padx=10)

        # ============ design frame_buttons ============

        # configure grid layout (3x2)
        self.equal_weighted_grid(self.frame_buttons, 3, 2)
        self.frame_buttons.grid_rowconfigure(0, minsize=10)
        self.frame_buttons.grid_rowconfigure(2, minsize=10)

        self.button_yes = customtkinter.CTkButton(
            master=self.frame_buttons,
            text='yes',
            font=('Consolas', 12),
            command=lambda: self.button_event('yes'),
        )
        self.button_yes.grid(row=1, column=0, pady=5, padx=20)

        self.button_no = customtkinter.CTkButton(
            master=self.frame_buttons,
            text='no',
            font=('Consolas', 12),
            command=lambda: self.button_event('no'),
        )
        self.button_no.grid(row=1, column=1, pady=5, padx=20)

        if self.countdown_seconds > 0:
            self.countdown()

        self.attributes('-topmost', True)
        self.mainloop()

    # __________________________________________________________
    # methods

    @staticmethod
    def equal_weighted_grid(obj: Any, rows: int, cols: int):
        """configures the grid to be of equal cell sizes for rows and columns."""
        for row in range(rows):
            obj.grid_rowconfigure(row, weight=1)
        for col in range(cols):
            obj.grid_columnconfigure(col, weight=1)

    def button_event(self, answer):
        """Stores the user input as instance attribute `answer`."""
        self.answer = answer
        self.terminate()

    def countdown(self):
        """Sets the timer for the question."""
        if self.answer is not None:
            self.terminate()
        elif self.remaining_seconds < 0:
            self.answer = self.default_value
            self.terminate()
        else:
            self.entry_timer.configure(textvariable=tkinter.StringVar(value=str(self.remaining_seconds)))
            self.remaining_seconds -= 1
            self.after(1000, self.countdown)

    def stop_after_callbacks(self):
        """Stops all after callbacks on the root."""
        for after_id in self.tk.eval('after info').split():
            self.after_cancel(after_id)

    def on_closing(self, event=0):
        """If the user presses the window x button without providing input"""
        if self.answer is None and self.default_value is not None:
            self.answer = self.default_value
        self.terminate()

    def terminate(self):
        """Properly terminates the gui."""
        if not self.terminated:
            # stop all .after callbacks to avoid error message "Invalid command ..." after destruction
            self.stop_after_callbacks()

            self.terminated = True
            try:
                self.destroy()
            except TclError:
                self.destroy()


# ______________________________________________________________________________________________________________________


"""
# example usage
if __name__ == '__main__':
    print('before')

    q1 = GuiPromptYesNo(question='1. do you want to proceed?', countdown_seconds=5)
    print(f'>>>{q1.answer=}')

    print('between')

    q2 = GuiPromptYesNo(question='2. do you want to proceed?', countdown_seconds=5)
    print(f'>>>{q2.answer=}')

    print('after')
"""


# ______________________________________________________________________________________________________________________



# db_functions.py
# -*- coding: utf-8 -*-

"""
Functions which serve for database purposes
"""

import warnings

from collections import namedtuple
import pandas as pd
from tqdm import tqdm

import sqlalchemy as sa
from sqlalchemy import exc

from .util_functions import input_prompt


# ______________________________________________________________________________________________________________________


def create_single_db_engine(db_cfg: dict, db_conn_info: dict, db_name: str = None) -> tuple:
    """
    Establishes an engine to a single database you can choose or specify upfront

    Args:
        db_cfg: dictionary containing the configuration information for the database
        db_conn_info: dictionary with the available database names as keys and connection information as values
        db_name: (optional) name of database, if None lets you choose from all available databases

    Returns:
        a named tuple with the following fields: db_name, engine
    """

    # db_cfg = cfg.Database.toDict(); db_conn_info = db_available; db_name = None;

    fields = ('db_name', 'engine')
    DB = namedtuple('DB', fields, defaults=(None,) * len(fields))

    if not db_name:
        db_name = input_prompt(
            name='database',
            choices=tuple(db_conn_info.keys()),
            enum=True,
        )

    conn_str = None
    if db_cfg[db_name]['type'] == 'MS SQL Server':
        conn_str = db_cfg[db_name]['driver'] + ':///?odbc_connect='
    elif db_cfg[db_name]['type'] in ('MariaDB', 'PostgreSQL', 'Oracle'):
        conn_str = db_cfg[db_name]['driver'] + '://'
    elif db_cfg[db_name]['type'] == 'SQLite3':
        conn_str = db_cfg[db_name]['driver'] + ':///'
    conn_str += str(db_conn_info[db_name])

    engine = sa.create_engine(conn_str)

    return DB(db_name=db_name, engine=engine)


def create_multiple_db_engines(db_cfg: dict, db_conn_info: dict, db_name: list[str] = None) -> dict:
    """
    Establishes one or multiple engines to databases you can choose or specify upfront

    Args:
        db_cfg: dictionary containing the configuration information for the database
        db_conn_info: dictionary with the available database names as keys and connection information as values
        db_name: (optional) name of database, if None lets you choose from all available databases

    Returns:
        a dict with the db_name as key and the engine as value
    """

    # db_cfg = cfg.Database.toDict(); db_conn_info = db_available; db_name = None;

    if not db_name or db_name == ['']:
        db_name = input_prompt(
            name='databases',
            message=(
                'to which dbs do you want to establish an engine'
                '(you can set up multiple, by comma separating the input)?'
            ),
            choices=tuple(db_conn_info.keys()) + ('all',),
            multi=True,
            enum=True,
        )

    db = {}

    if 'all' in db_name:
        for name in db_conn_info.keys():
            _, db[name] = create_single_db_engine(db_cfg, db_conn_info, db_name=name)
    else:
        for name in db_name:
            _, db[name] = create_single_db_engine(db_cfg, db_conn_info, db_name=name)

    return db


# ______________________________________________________________________________________________________________________


def execute_raw_sql(qry: str | sa.sql.elements.TextClause, con: sa.engine.Connection | sa.engine.Engine) -> None:
    """
    Executes a sql statement and does not return anything

    Args:
        qry: the sql query to be executed, either a string or sqlalchemy text clause
        con: either the sqlalchemy connection or engine to the database

    Returns:
        Executes and commits the statement against the database and returns None
    """

    if type(qry).__name__ == 'str':
        qry = sa.text(qry)
    elif type(qry).__name__ != 'TextClause':
        raise TypeError('argument qry must be either a string or sqlalchemy.text!')

    con_type = type(con).__name__
    if con_type == 'Engine':
        con = con.connect()
    try:
        con.execute(qry)
        con.commit()
    except Exception as ex:
        raise ex
    finally:
        if con_type == 'Engine':
            con.close()

    return None


# ______________________________________________________________________________________________________________________


def loop_insert_df_to_table(
        df: pd.DataFrame,
        tbl_name: str,
        tbl_schema: str,
        con: sa.engine.Connection | sa.engine.Engine,
        if_exists: str = 'append',
        verbose: bool = False,
) -> list:
    """
    inserts rows to a database table in a loop

    Args:
        df: the dataframe which should be inserted
        tbl_name: the name of the database table
        tbl_schema: the name of the database schema where the table is located
        con: either the sqlalchemy connection or engine to the database
        if_exists: what to do if the table already exists
        verbose: whether to show a loop progress bar and print statements

    Returns:
        a list with the indexes of the rows which where NOT inserted due to integrity errors
    """

    dup_rows = []
    len_df = len(df)
    df.reset_index(drop=True, inplace=True)
    sequence = tqdm(range(len_df)) if verbose else range(len_df)
    for i in sequence:
        try:
            df.loc[i:i].to_sql(tbl_name, schema=tbl_schema, index=False, con=con, if_exists=if_exists, method=None)
        except exc.IntegrityError:
            dup_rows.append(i)
    if len(dup_rows) > 0:
        warnings.warn(f'{len(dup_rows)} integrity error(s) due to duplicate row(s)')
    if verbose:
        stmt = f'{len_df - len(dup_rows)} row(s) inserted'
        stmt += f', {len(dup_rows)} duplicate row(s) where not inserted' if len(dup_rows) > 0 else ''
        print(stmt)

    return dup_rows

# ______________________________________________________________________________________________________________________



# util_decorators.py
# -*- coding: utf-8 -*-

"""
Decorators which serve for general purposes
"""

import functools


# ______________________________________________________________________________________________________________________


def timed(reps: int = 1):
    """
    decorator factory function which creates a decorator to measure execution time of functions

    Args:
        reps: how many times to repeat executing the function

    Returns:
        a decorator function
    """

    def dec(fn):
        """
        decorator

        Args:
            fn: the function to be decorated

        Returns:
            decorated function
        """

        from time import perf_counter

        @functools.wraps(fn)
        def inner(*args, **kwargs):
            """
            closure which runs the decorator code and executes the function

            Args:
                args: arbitrary number of positional arguments
                kwargs: arbitrary number of keyword only arguments

            Returns:
                the result of the function call to fn
            """

            fn_result = None
            arguments = ', '.join([str(a) for a in args])
            kw_arguments = ', '.join([f'{k}={v}' for k, v in kwargs.items()])
            all_arguments = ', '.join(filter(None, [arguments, kw_arguments]))

            total_elapsed = 0

            for i in range(reps):
                start = perf_counter()
                fn_result = fn(*args, **kwargs)
                total_elapsed += (perf_counter() - start)
            avg_elapsed = total_elapsed / reps

            if reps == 1:
                print(f'{fn.__name__}({all_arguments}) execution time : {avg_elapsed}')
            else:
                print(f'{fn.__name__}({all_arguments}) avg execution time after {reps} reps: {avg_elapsed}')
            return fn_result

        return inner

    return dec


# ______________________________________________________________________________________________________________________



# util_context_managers.py
# -*- coding: utf-8 -*-

"""
Context managers which serve for general purposes
"""

from time import perf_counter


# ______________________________________________________________________________________________________________________


class Timer:
    """
    A context manager to measure start, end and elapsed time of a code block.
    Useage:

    >>> with Timer() as time_stats:
    >>>     '...write code here..'
    >>> print(time_stats.elapsed)
    """

    def __init__(self):
        self.elapsed = 0

    def __enter__(self):
        self.start = perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop = perf_counter()
        self.elapsed = self.stop - self.start
        return False


# ______________________________________________________________________________________________________________________



# data_functions.py
# -*- coding: utf-8 -*-

"""
Functions which serve for data transformation purposes
"""

import itertools

from tqdm import tqdm

import pandas as pd


# ______________________________________________________________________________________________________________________


def stack_temporal_dataframe(
        df: pd.DataFrame,
        start_col: str,
        end_col: str,
        freq: str,
        silent: bool = True,
) -> pd.DataFrame:
    """
    Transforms a dataframe along a date range,
    from columns [start_col, end_col] to one column [temporal_range].

    Args:
        df:         Input dataframe. Must be unique in rows!
        start_col:  Name of the start date column
        end_col:    Name of the end date column
        freq:       Interval frequency. Use a pandas DateOffset object from ->
                    https://pandas.pydata.org/docs/user_guide/timeseries.html#dateoffset-objects
        silent:     If False, a tqdm progres bar will be shown

    Returns:
        transformed dataframe with new column 'temporal_range'
    """

    records = []
    iterator = df.to_dict(orient='records')
    iterator = iterator if silent else tqdm(iterator)

    for row in iterator:
        temporal_range = pd.date_range(row[start_col], row[end_col], freq=freq)

        l_tmp = list(
            itertools.product(
                [[el] for el in temporal_range],
                [list(row.values())],
            )
        )

        records.extend(list(map(lambda x: sum(x, []), l_tmp)))

    df_output = pd.DataFrame(records, columns=['temporal_records'] + list(df.columns))
    df_output.drop(columns=[start_col, end_col], inplace=True)
    df_output.reset_index(drop=True, inplace=True)
    # recast to original dtypes from input df
    for col, dtype in df.dtypes.items():
        if col in df_output.columns and col != 'temporal_column' and df_output[col].dtype != dtype:
            df_output[col] = df_output[col].astype(dtype)

    return df_output


# ______________________________________________________________________________________________________________________


def unstack_temporal_dataframe(
        df: pd.DataFrame,
        temporal_column: str,
        group_by: list,
        silent: bool = True,
) -> pd.DataFrame:
    """
    Transforms a dataframe along a date range,
    from one column [temporal_range] to columns [start_col, end_col].

    Args:
        df:                 Input dataframe
        temporal_column:    Name of temporal column
        group_by:           Name of columns for which to group by
        silent:             If False, a tqdm progres bar will be shown

    Returns:
        transformed dataframe with new columns 'valid_from' and 'valid_to'
    """
    if isinstance(group_by, str):
        group_by = [group_by]

    iterator = df.groupby(by=group_by)
    iterator = iterator if silent else tqdm(iterator)

    rows_output = []
    for _, group in iterator:
        temporal_index = group[temporal_column].to_list()
        group.set_index(keys=[temporal_column] + group_by, inplace=True)

        comparison = group.compare(group.shift())

        rows = []

        # special case: no delta in complete group
        if len(comparison) == 1:
            row = {
                'valid_from': temporal_index[0],
                'valid_to': temporal_index[-1],
            }

            for k in group_by:
                row[k] = group.reset_index()[k][0]

            for k in group.columns:
                row[k] = group.reset_index()[k][0]

            rows.append(row)

        else:
            for i, idx in enumerate(comparison.index):
                # special case: first entry
                if i == 0:
                    continue
                else:
                    valid_from = comparison.iloc[i - 1].name[0]
                    valid_to = temporal_index[temporal_index.index(idx[0]) - 1]

                row = {
                    'valid_from': valid_from,
                    'valid_to': valid_to,
                }

                for j, k in enumerate(group_by):
                    row[k] = idx[j + 1]

                for k in group.columns:
                    row[k] = group.shift().loc[idx][k]

                rows.append(row)

            # special case: last entry
            row = {
                'valid_from': idx[0],
                'valid_to': group.index[-1][0],
            }

            for j, k in enumerate(group_by):
                row[k] = idx[j + 1]

            for k in group.columns:
                row[k] = group.loc[idx][k]

            rows.append(row)

        rows_output.extend(rows)

    df_output = pd.DataFrame.from_records(rows_output)
    df_output.reset_index(drop=True, inplace=True)
    # recast to original dtypes from input df
    for col, dtype in df.dtypes.items():
        if col in df_output.columns and col != temporal_column and df_output[col].dtype != dtype:
            df_output[col] = df_output[col].astype(dtype)

    return df_output


# ______________________________________________________________________________________________________________________



# util_classes.py
# -*- coding: utf-8 -*-

"""
Classes which serve for general purposes
"""

import time

import logging

import ssl
import smtplib
from email.message import EmailMessage


# ______________________________________________________________________________________________________________________


class StartTime:
    called = False

    @classmethod
    def show(cls, logger: logging.Logger) -> None:
        if not cls.called:
            logger.info('____________________________________________________________')
            logger.info(f'>> Execution started at {time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime())} <<')
            cls.called = True
        return None


# ______________________________________________________________________________________________________________________


class Email:
    def __init__(self, server: str, port: int, address: str, password: str | None, logger: logging.Logger):
        self.server = server
        self.port = port
        self.address = address
        self.password = password
        self.logger = logger

    def send_email(
            self, mail_to: str | list, subject: str, body: str, html: bool = False, use_ssl: bool = True
    ) -> None:
        """
        sending the email

        Args:
            mail_to: receiver, can be a list of strings
            subject: the subject of the email
            body: the body of the email
            html: whether the body is formatted with html or not
            use_ssl: whether to use a secure ssl connection and authentication

        Returns:
            None
        """
        if not isinstance(mail_to, str):
            mail_to = ', '.join(mail_to)

        if self.server == 'localhost':
            with smtplib.SMTP(self.server, self.port) as server:
                message = f'Subject: {subject}\n\n{body}'
                try:
                    server.sendmail(self.address, mail_to, message)
                    return None
                except smtplib.SMTPException as ex:
                    self.logger.exception('an error occurred!')
                    raise ex
        else:
            mail = EmailMessage()
            mail['Subject'] = subject
            mail['From'] = self.address
            mail['To'] = mail_to
            if html:
                mail.add_alternative(body, subtype='html')
            else:
                mail.set_content(body)

        if use_ssl:
            with smtplib.SMTP_SSL(self.server, self.port, context=ssl.create_default_context()) as server:
                server.login(self.address, self.password)
                server.send_message(mail)
        else:
            with smtplib.SMTP(self.server, self.port) as server:
                server.send_message(mail)

        return None


# ______________________________________________________________________________________________________________________
